import re
from typing import Optional

import discord


bot: Optional[discord.Client] = None


def pass_bot(_bot):
    global bot
    bot = _bot


class ParseMention:
    REGEX_ROLES = r"&lt;@&amp;([0-9]+)&gt;"
    REGEX_MEMBERS = r"&lt;@!?([0-9]+)&gt;"
    REGEX_CHANNELS = r"&lt;#([0-9]+)&gt;"
    REGEX_EMOJIS = r"&lt;a?(:[^\n:]+:)[0-9]+&gt;"
    REGEX_ROLES_2 = r"<@&([0-9]+)>"
    REGEX_MEMBERS_2 = r"<@!?([0-9]+)>"
    REGEX_CHANNELS_2 = r"<#([0-9]+)>"
    REGEX_EMOJIS_2 = r"<a?(:[^\n:]+:)[0-9]+>"

    ESCAPE_LT = "______lt______"
    ESCAPE_GT = "______gt______"
    ESCAPE_AMP = "______amp______"

    def __init__(self, content, guild):
        self.content = content
        self.guild = guild

    def flow(self):
        self.escape_mentions()
        self.escape_mentions()
        self.unescape_mentions()
        self.normal_channel_mention()
        self.html_channel_mention()
        self.normal_user_mention()
        self.html_user_mention()
        self.normal_role_mention()
        self.html_role_mention()

        return self.content

    def escape_mentions(self):
        for match in re.finditer("(%s|%s|%s|%s|%s|%s|%s|%s)"
                                 % (self.REGEX_ROLES, self.REGEX_MEMBERS, self.REGEX_CHANNELS, self.REGEX_EMOJIS,
                                    self.REGEX_ROLES_2, self.REGEX_MEMBERS_2, self.REGEX_CHANNELS_2,
                                    self.REGEX_EMOJIS_2), self.content):
            pre_content = self.content[:match.start()]
            post_content = self.content[match.end():]
            match_content = self.content[match.start():match.end()]

            match_content = match_content.replace("<", self.ESCAPE_LT)
            match_content = match_content.replace(">", self.ESCAPE_GT)
            match_content = match_content.replace("&", self.ESCAPE_AMP)

            self.content = pre_content + match_content + post_content

    def unescape_mentions(self):
        self.content = self.content.replace(self.ESCAPE_LT, "<")
        self.content = self.content.replace(self.ESCAPE_GT, ">")
        self.content = self.content.replace(self.ESCAPE_AMP, "&")
        pass

    def normal_channel_mention(self):
        offset = 0
        for match in re.finditer(self.REGEX_CHANNELS, self.content):
            channel_id = int(match.group(1))
            channel = self.guild.get_channel(channel_id)
            if channel is None:
                replacement = '#deleted-channel'
            else:
                replacement = '<span class="mention" title="%s">#%s</span>' \
                              % (channel.name, channel.name)
            self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset],
                                                replacement)
            offset += len(replacement) - (match.end() - match.start())

    def html_channel_mention(self):
        offset = 0
        for match in re.finditer(self.REGEX_CHANNELS_2, self.content):
            channel_id = int(match.group(1))
            channel = self.guild.get_channel(channel_id)
            if channel is None:
                replacement = '#deleted-channel'
            else:
                replacement = '<span class="mention" title="%s">#%s</span>' \
                              % (channel.name, channel.name)
            self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset],
                                                replacement)
            offset += len(replacement) - (match.end() - match.start())

    def normal_role_mention(self):
        offset = 0
        for match in re.finditer(self.REGEX_ROLES, self.content):
            role_id = int(match.group(1))
            role = self.guild.get_role(role_id)
            if role is None:
                replacement = '@deleted-role'
            else:
                replacement = '<span style="color: #%02x%02x%02x;">@%s</span>' \
                              % (role.color.r, role.color.g, role.color.b, role.name)
            self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset], replacement)
            offset += len(replacement) - (match.end() - match.start())

    def html_role_mention(self):
        offset = 0
        for match in re.finditer(self.REGEX_ROLES_2, self.content):
            role_id = int(match.group(1))
            role = self.guild.get_role(role_id)
            if role is None:
                replacement = '@deleted-role'
            else:
                replacement = '<span style="color: #%02x%02x%02x;">@%s</span>' \
                              % (role.color.r, role.color.g, role.color.b, role.name)
            self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset], replacement)
            offset += len(replacement) - (match.end() - match.start())

    def normal_user_mention(self):
        offset = 0
        for match in re.finditer(self.REGEX_MEMBERS, self.content):
            member_id = int(match.group(1))

            member = None
            try:
                member = self.guild.get_member(member_id) or bot.get_user(member_id)
                member_name = member.display_name
            except AttributeError:
                member_name = member

            if member is not None:
                replacement = '<span class="mention" title="%s">@%s</span>' \
                              % (member, str(member_name))
                self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset],
                                                    replacement)
                offset += len(replacement) - (match.end() - match.start())
            else:
                self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset],
                                                    '<span class="mention" title="%s">@%s</span>' % (
                                                        "Unknown", "Unknown"))

    def html_user_mention(self):
        offset = 0
        for match in re.finditer(self.REGEX_MEMBERS_2, self.content):
            member_id = int(match.group(1))

            member = None
            try:
                member = self.guild.get_member(member_id) or bot.get_user(member_id)
                member_name = member.display_name
            except AttributeError:
                member_name = member

            if member is not None:
                replacement = '<span class="mention" title="%s">@%s</span>' \
                              % (member, str(member_name))
                self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset],
                                                    replacement)
                offset += len(replacement) - (match.end() - match.start())
            else:
                self.content = self.content.replace(self.content[match.start() + offset:match.end() + offset],
                                                    '<span class="mention" title="%s">@%s</span>' % (
                                                        "Unknown", "Unknown"))



import math

from chat_exporter.build_html import fill_out, img_attachment, msg_attachment, audio_attachment, PARSE_MODE_NONE


class BuildAttachment:
    image_types = ".png", ".jpeg", ".jpg", ".gif"
    audio: str = ""
    file_icon: str = ""

    def __init__(self, attachments, guild):
        self.attachments = attachments
        self.guild = guild

    async def flow(self):
        await self.build_attachment()

        return self.attachments

    async def build_attachment(self):
        if str(self.attachments.url).endswith(self.image_types):
            await self.image()
        else:
            await self.file()

    async def image(self):
        self.attachments = await fill_out(self.guild, img_attachment, [
            ("ATTACH_URL", self.attachments.proxy_url, PARSE_MODE_NONE),
            ("ATTACH_URL_THUMB", self.attachments.proxy_url, PARSE_MODE_NONE)
        ])

    async def file(self):
        await self.get_file_icon()

        file_size = self.get_file_size(self.attachments.size)

        self.attachments = await fill_out(self.guild, msg_attachment, [
            ("ATTACH_ICON", self.file_icon, PARSE_MODE_NONE),
            ("ATTACH_URL", self.attachments.url, PARSE_MODE_NONE),
            ("ATTACH_BYTES", str(file_size), PARSE_MODE_NONE),
            ("ATTACH_AUDIO", self.audio, PARSE_MODE_NONE),
            ("ATTACH_FILE", str(self.attachments.filename), PARSE_MODE_NONE)
        ])

    @staticmethod
    def get_file_size(file_size):
        if file_size == 0:
            return "0 bytes"
        size_name = ("bytes", "KB", "MB")
        i = int(math.floor(math.log(file_size, 1024)))
        p = math.pow(1024, i)
        s = round(file_size / p, 2)
        return "%s %s" % (s, size_name[i])

    async def get_file_icon(self):
        audio_types = "aac", "mid", "mp3", "m4a", "ogg", "flac", "wav", "amr"
        webcode_types = "html", "htm", "css", "rss", "xhtml", "xml"
        code_types = "py", "cgi", "pl", "gadget", "jar", "msi", "wsf", "bat", "php", "js"
        document_types = "txt", "doc", "docx", "rtf", "xls", "xlsx", "ppt", "pptx", "odt", "odp", "ods", "odg", "odf", \
                         "swx", "sxi", "sxc", "sxd", "stw"
        acrobat_types = "pdf"
        archive_types = "br", "rpm", "dcm", "epub", "zip", "tar", "rar", "gz", "bz2", "7x", "deb", "ar", "Z", "lzo", \
                        "lz", "lz4", "arj", "pkg", "z"

        split_url = self.attachments.url.rsplit('.', 1)

        if split_url[1].endswith(audio_types):
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-audio.svg"
            await self.set_audio()
        elif split_url[1].endswith(webcode_types):
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-webcode.svg"
        elif split_url[1].endswith(code_types):
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-code.svg"
        elif split_url[1].endswith(document_types):
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-document.svg"
        elif split_url[1].endswith(acrobat_types):
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-acrobat.svg"
        elif split_url[1].endswith(archive_types):
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-archive.svg"
        else:
            self.file_icon = "https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/discord-unknown.svg"

    async def set_audio(self):
        self.audio = await fill_out(self.guild, audio_attachment, [
            ("ATTACH_URL", self.attachments.url, PARSE_MODE_NONE),
            ("ATTACH_FILE", str(self.attachments.filename))
        ])



##################################################################################
# Copyright (c) 2016, Katie McLaughlin                                           #
# All rights reserved.                                                           #
#                                                                                #
# Redistribution and use in source and binary forms, with or without             #
# modification, are permitted provided that the following conditions are met:    #
#                                                                                #
# * Redistributions of source code must retain the above copyright notice, this  #
#   list of conditions and the following disclaimer.                             #
#                                                                                #
# * Redistributions in binary form must reproduce the above copyright notice,    #
#   this list of conditions and the following disclaimer in the documentation    #
#   and/or other materials provided with the distribution.                       #
#                                                                                #
# * Neither the name of octohatrack nor the names of its                         #
#   contributors may be used to endorse or promote products derived from         #
#   this software without specific prior written permission.                     #
#                                                                                #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"    #
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE      #
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE   #
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL     #
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR     #
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER     #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  #
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  #
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.           #
#                                                                                #
# Github: https://github.com/glasnt/emojificate                                  #
##################################################################################


import unicodedata
from grapheme import graphemes
import emoji
import requests


__all__ = ["convert_emoji"]

cdn_fmt = "https://twemoji.maxcdn.com/v/latest/72x72/{codepoint}.png"


def valid_src(src):
    req = requests.head(src)
    return req.status_code == 200


def valid_category(char):
    try:
        return unicodedata.category(char) == "So"
    except TypeError:
        return False


def get_best_name(char):
    """
    unicode data does not recognise the grapheme,
    so try and parse something from emoji instead.
    """
    shortcode = emoji.demojize(char, use_aliases=True)

    # Roughly convert shortcode to screenreader-friendly sentence.
    return shortcode.replace(":", "").replace("_", " ").replace("selector", "").title()


def convert(char):
    def tag(a, b):
        return '%s="%s"' % (a, b)

    def codepoint(codes):
        # See https://github.com/twitter/twemoji/issues/419#issuecomment-637360325
        if "200d" not in codes:
            return "-".join([c for c in codes if c != "fe0f"])
        return "-".join(codes)

    if valid_category(char):
        # Is a Char, and a Symbol
        name = unicodedata.name(char).title()
    else:
        if len(char) == 1:
            # Is a Char, not a Symbol, we don't care.
            return char
        else:
            # Is probably a grapheme
            name = get_best_name(char)

    src = cdn_fmt.format(codepoint=codepoint(["{cp:x}".format(cp=ord(c)) for c in char]))

    # If twitter doesn't have an image for it, pretend it's not an emoji.
    if valid_src(src):
        return "".join(
            [
                '<img class="emoji emoji--small" ',
                tag(" src", src),
                tag(" alt", char),
                tag(" title", name),
                tag(" aria-label", "Emoji: %s" % name),
                ">",
            ]
        )
    else:
        return char


def convert_emoji(string):
    return "".join(convert(ch) for ch in graphemes(string))
